-------------------------------------------------------------------------------
-- Fortify
--
-- A modded version of Decursive used to buff everyone in your party automatically.
-- Created by Pravetz (Proudmoore)
--
-- Date      Rev    Description
-- --------  -----  -----------------------
-- 06/12/05  0.5    Initial version of Fortify casts Inner Fire, Fortitude, Divine Spirit,
--                  and Shadow Protection.
--                  Added Feedback (weapon buff).
--                  Casts the appropriate ranks of spells for lower level chars.
--                  Check to see that we actually have the spells before casting.
--                  Corrected unusual targeting error messages.
-- 06/13/05  0.7    Add Elune's Grace for night elf priests.
--                  Only cast divine spirit on people who have mana.
-- 06/15/05  1.0    Fix this wierd range thing when people are in different zones.
--                  Added clearer party/raid-announce messages.
--                  Released mod.
-- 06/18/05  1.1    Trimmed out some excess code and removed some unused logic.
--                  Sped up the spell selection routines.  It does fewer searches through
--                  the spellbook now.
--                  Added message to indicate nothing to do.
--                  Added a user interface button for quick use.
-- 06/19/05  1.2    Added a "quiet" command for SilvioG.  Type "/fortify quiet".
--                  Added a texture for the user interface button.
--                  Moved all the speaking stuff into a separate function.
--                  Sped up the range checking (check once per unit instead of once per spell).
-- 06/20/05         Added a "move" command so the icon can be relocated.  Doesn't work yet.
--                  Check to see if the player is mounted.
--                  Allow the player to turn off certain buffs if not desired.
-- 06/21/05  1.3    Corrected the bad version 1.2 that I posted.
--                  Added "just" and "all" commands.
--                  Added "show" to show all current settings.
--                  Remembers your settings.
--                  Added Paladin blessings.
--                  Use "whispercast" style requests to let users select their paladin blessings.
--                  Added Troll racial spell Shadowguard.
-- 06/28/05  1.4    Added user interface.  Properly movable.
--                  Shows Decursive button as well.
--                  Ignores dead and offline units.
-- 06/28/05  1.5    Added "Mark of the Wild" and "Thorns" - apparently I forgot to include them in BuffUnit.
--                  Moved all spell data into data.lua.
-- 08/05/05  1.6    Revised code to use better data structures.
--                  Added an "auto-reply" toggle to disable the automatic replies.
--                  There are three reply modes: auto, fail, and off.
--                  Improved the range checking code - at least, it won't get stuck on people.
-- 08/07/05  1.7    Corrected a problem with spell overrides.
--                  Also, typing "/ffy all" will correct any bugs with misconfigured spell data.
--                  Returned the paladin code back into place - sorry, guys.
--                  /ffy toggle now uses shortcuts to identify spells.
-- 08/10/05  1.8    Yet another problem with spell overrides.
--                  A little more work on Paladin blessings.
-- 08/17/05  1.9    Re-tested and validated paladin blessings.
--                  Added (optional) pet buffs.
--                  Corrected problem with whispering when you have no buff spells to cast.
--                  Fixed the error in paladin blessings that was causing it to lose track of preferred blessings.
--                  Added "party blessings" flag to properly handle salvation.
--                  Added optional tell notification.
-- 08/21/05  1.91   Corrected minor typo.
-- 09/05/05  2.00   Added UI icon checking for range.
--                  Added Mage self-cast spells.
--                  Added improved re-attack after fortify is cast.
--                  Added /ffy announce command.
--                  Added /ffy bless selection for Paladins.
--                  Added demon skin for warlocks.
--                  Added PVP detection to not buff people if they'd alter your PVP status
--                  Added recast when the buff is about to expire - it's fixed at 30 secs till expiration
--                  But, it won't recast if the buff is still on the person after the 30 secs are up, in
--                  case someone else cast it on the target.
--                  The "/ffy Recast" command will now force all buffs to recast - before a boss, etc.
-- 01/01/06  2.1    Buff selects Kings as the first preference for non-casters.
--                  Hunters get Wisdom before Kings.
--                  Added testing for dropped buffs every 30 seconds.
--                  Flag the user interface to indicate some buffs need to be cast.
--                  "Stop-at" for mana conservation.
--                  Only attack the unit specified if they are attackable (don't fight dead units)
-- 02/20/06  2.2    Fixed bug with Paladin blessings.  They cast now.
--                  Added french translations from mymycracra (Thank you!)
--                  
--
-- To Do:
-- Calculate total mana costs for all buffs.
-- Add alt-self-cast so I can get rid of easypriest.
-- Add auto-drink out of combat.
-- Check buff levels when /ffy toggle is used
-- Update buff status as soon as spells are cast
-- Separate buffs into cast-in-combat vs no cast in combat
-- Allow /ffy recast for individual spells
--


-------------------------------------------------------------------------------

-- This keeps track of all the units we are monitoring
local ffy_Unit_Array        = { };
local ffy_LastUnitChecked   = 0;

-- used to test for posession of a spell and its index in the spell system
local ffy_SpellID           = nil;
local ffy_SpellIconSlot     = nil;
local ffy_RecastTimes       = nil;

-- Debug flags - set them to true if you want garbage spam
local ffy_Print_DEBUG       = false;

-- These are all the ffy_Options that are saved on logout
ffy_Options = {
    Version = "unknown",
	NotifyGroup = true,
	ShowPanel = true,
	ReplyWhisper = FFY_REPLY_AUTO,
	Enabled = { },
	PanelMovable = true,
	CastOnPets = false,
	TellBuffs = false,
	DutyCycle = FFY_DUTY_ALL,
	ManaConserve = 4000,
	RescanTimer = 10.0,
};

-- Have we started or finished casting?
local ffy_Is_Buffing        = false;
local ffy_People_Missed     = "";

-- What paladin blessings have we cast on people?
local ffy_Paladin_Blessings = { };

-- When did we last cast a buff on someone?
local ffy_NextBuffTime = { };
local ffy_RecastList = { };

-- This is the overall list we find by scanning your spellbook
local ffy_MyData = nil;
local ffy_SpellRanks = { };

-- Game Constants
FFY_MAXBUFFS   = 16;
FFY_START_SLOT = 1;
FFY_END_SLOT   = 120;

-- A couple useful variables we track
ffy_AttackOnDelay   = 0.0;
ffy_InCombatMode    = false;
ffy_IsPaladin       = false;
ffy_Clock           = 0.0;
ffy_RecastTimer     = 30.0;
ffy_FullRecastMode  = false;
ffy_NextRescan      = 30.0;

-------------------------------------------------------------------------------
-- and the printing functions
-------------------------------------------------------------------------------
function ffy_debug(Message)
    if (ffy_Print_DEBUG) then
        DEFAULT_CHAT_FRAME:AddMessage(Message, 0.1, 0.1, 1);
    end
end

function ffy_println(Message)
    DEFAULT_CHAT_FRAME:AddMessage(FFY_TAG .. Message, 1, 1, 1);
end

function ffy_print(Message)
    DEFAULT_CHAT_FRAME:AddMessage(Message, 1, 1, 1);
end

function ffy_errln(Message)
    DEFAULT_CHAT_FRAME:AddMessage(FFY_TAG .. Message, 1, 0.1, 0.1);
end

function ffy_speak(Message)
    if (ffy_Options.NotifyGroup) then
        ffy_forcespeak(Message);
    else
        ffy_println(Message);
    end
end

function ffy_forcespeak(Message)
    if (GetNumRaidMembers() > 0) then
        SendChatMessage(FFY_TAG .. Message, 'raid');
    elseif (GetNumPartyMembers() > 0) then
        SendChatMessage(FFY_TAG .. Message, 'party');
    else
        ffy_println(Message);
    end
end

-------------------------------------------------------------------------------
-- Start up Fortify
-------------------------------------------------------------------------------
function ffy_Init()

    -- Announce ourselves
    ffy_println(FFY_VERSION_STRING);

    -- Register our slash commands
    SlashCmdList["FORTIFY"] = ffy_Commands;
    SLASH_FORTIFY1 = FFY_MACRO_COMMAND;
    SLASH_FORTIFY2 = FFY_MACRO_COMMAND2;
    
    -- Register all the events we handle
    this:RegisterEvent("CHAT_MSG_WHISPER");
    this:RegisterEvent("SPELLS_CHANGED");

    ffy_Configure();

    -- create the array of curable items
    ffy_CreateTargetList();
end

-------------------------------------------------------------------------------
-- Create the target list based on the current duty cycle
-------------------------------------------------------------------------------
function ffy_OnUpdate(arg1)

	-- Imitated from Decursive (right on quu)
	if (ffy_AttackOnDelay > 0) then
		ffy_AttackOnDelay = ffy_AttackOnDelay - arg1;
		if (ffy_AttackOnDelay <= 0) then
			if (not ffy_InCombatMode) then
			    if (UnitCanAttack("player", "target")) then
    				AttackTarget();
                    ffy_debug("Attacking on delay.");
                end
			end
		end
	end
	
	-- Check if it's time to run a scan - once every 30 seconds
	if (ffy_NextRescan < ffy_Clock) then
	    ffy_Rescan();
	    ffy_NextRescan = ffy_NextRescan + ffy_Options.RescanTimer;
	end
	
	-- Update the fortify clock
	ffy_Clock = ffy_Clock + arg1	
end


-------------------------------------------------------------------------------
-- Create the target list based on the current duty cycle
-------------------------------------------------------------------------------
function ffy_Rescan()
    if (ffy_AnyBuffsToCast()) then
        ffy_Button:SetAlpha(1.0);
        ffy_debug("Time to cast Fortify!");
    else
        ffy_Button:SetAlpha(0.25);
        ffy_debug("No buffs needed now.");
    end
end


-------------------------------------------------------------------------------
-- Create the target list based on the current duty cycle
-------------------------------------------------------------------------------
function ffy_CreateTargetList()
    local i;
    
    -- Re-Initialize the unit array
    ffy_Unit_Array = { };
    
    -- Add yourself first
    table.insert(ffy_Unit_Array, "player");
    if (ffy_Options.CastOnPets) then
        table.insert(ffy_Unit_Array, "pet");
    end
    
    -- Then your target
    table.insert(ffy_Unit_Array, "target");
    
    -- Then your party
    for i = 1, 4 do
        table.insert(ffy_Unit_Array, "party"..i);
        if (ffy_Options.CastOnPets) then
            table.insert(ffy_Unit_Array, "partypet"..i);
        end
    end
    
    -- Then the whole raid
    for i = 1, 40 do
        table.insert(ffy_Unit_Array, "raid"..i);
        if (ffy_Options.CastOnPets) then
            table.insert(ffy_Unit_Array, "raidpet"..i);
        end
    end
end

-------------------------------------------------------------------------------
-- Fortify text user interface
-------------------------------------------------------------------------------
function ffy_Commands(msg)

    -- The command is the next text off the message
    local command, args, p;
    p = string.find(msg," ");
    if (p ~= nil) then
        args = string.sub(msg,p + 1);
        command = string.lower(string.sub(msg,1,p - 1));
        ffy_debug("Selected [" .. command .. "] [" .. args .. "]");
    else
        command = string.lower(msg);
        ffy_debug("Selected [" .. command .. "]");
    end
    
    -- If fortify is done as a macro
    if (command == "") then
        ffy_Fortify();
        
    -- Toggle group notifications
    elseif (command == FFY_CMD_QUIET) then
        ffy_Options.NotifyGroup = not ffy_Options.NotifyGroup;
        if (ffy_Options.NotifyGroup) then
            ffy_println(FFY_MODE_NOTIFY);
        else
            ffy_println(FFY_MODE_QUIET);
        end
        
    -- Change mana conservation levels
    elseif (command == FFY_CMD_CONSERVE) then
        ffy_Options.ManaConserve = tonumber(args);
        if (ffy_Options.ManaConserve == nil) or (ffy_Options.ManaConserve < 0) then
            ffy_Options.ManaConserve = 0;
        end
        if (ffy_Options.ManaConserve > UnitManaMax("player")) then
            ffy_Options.ManaConserve = UnitManaMax("player");
        end
        ffy_println(string.gsub(FFY_CONSERVE_MSG, "$m", ffy_Options.ManaConserve));
        ffy_Rescan();
        
    -- Toggle responses to whispercast requests
    elseif (command == FFY_CMD_REPLY) then
        if (args == FFY_REPLY_AUTO) then
            ffy_Options.ReplyWhisper = args;
            ffy_println(FFY_MODE_AUTOREPLY);
        elseif (args == FFY_REPLY_FAIL) then
            ffy_Options.ReplyWhisper = args;
            ffy_println(FFY_MODE_FAILREPLY);
        elseif (args == FFY_REPLY_OFF) then
            ffy_Options.ReplyWhisper = args;
            ffy_println(FFY_MODE_NOREPLY);
        else
            ffy_println(FFY_REPLY_CHOICES);
        end
        
    -- Show or hide the UI panel
    elseif (command == FFY_CMD_TOGGLEUI) then
        ffy_Options.ShowPanel = not ffy_Options.ShowPanel;
        ffy_Button_Update();
        
    -- Make the buff icon moveable (DOESN"T WORK YET)
    elseif (command == FFY_CMD_MOVEABLE) then
        
        ffy_Options.PanelMovable = not ffy_Options.PanelMovable;
        ffy_ButtonFrame.movable = ffy_Options.PanelMovable;
        if (ffy_Options.PanelMovable) then
            ffy_println("The Fortify button is now movable.  Click and drag it to move it.");
        else
            ffy_println("The Fortify button is locked in place.");
        end
        
    -- Toggle individual spells notifications
    elseif (command == FFY_CMD_TOGGLESPELL) then
        if (args == nil) then
            ffy_println(FFY_TOGGLE_SYNTAX);
        else
            local spellname = ffy_FindSpellByName(args);
            if (spellname == nil) then
                ffy_println(string.gsub(FFY_TOGGLE_SPELLING, "$s", args));
            else
                ffy_Options.Enabled[spellname] = not ffy_Options.Enabled[spellname];
                if (ffy_Options.Enabled[spellname]) then
                    ffy_println(string.gsub(FFY_TOGGLE_YESCAST, "$s", spellname));
                else
                    ffy_println(string.gsub(FFY_TOGGLE_NOCAST, "$s", spellname));
                end
            end
        end
        ffy_Rescan();
        
    -- Toggle individual spells notifications
    elseif (command == FFY_CMD_SINGLESPELL) then
        local spellname, spellstate, onlyspell;
        
        -- If the spell name can be found, turn it on and all else off
        onlyspell = ffy_FindSpellByName(args)
        if (onlyspell) then
            for spellname, settings in ffy_MyData do
                ffy_Options.Enabled[spellname] = (spellname == onlyspell);
            end
            ffy_println(string.gsub(FFY_TOGGLE_SINGLE, "$s", onlyspell));
            
        -- Otherwise print an error message
        else
            ffy_println(FFY_ONLY_SYNTAX);
        end
        ffy_Rescan();
    
    -- Toggle individual spells notifications
    elseif (command == FFY_CMD_ALLSPELLS) then
        local playerClass, englishClass = UnitClass("player");
        ffy_MyData = FFY_SPELL_DATA[englishClass];
        for spellname, settings in ffy_MyData do
            ffy_Options.Enabled[spellname] = true;
        end
        ffy_println(FFY_TOGGLE_ALL);
        ffy_Rescan();
    
    -- Toggle individual spells notifications
    elseif (command == FFY_CMD_SHOW) then
        local playername, spellname, spellstate, showline, ignoreline
        ffy_println(FFY_SHOW_SETTINGS);
        showline = "";
        ignoreline = "";
        for spellname, settings in ffy_MyData do
            if (ffy_SpellRanks[spellname] > 0) then
                if (ffy_Options.Enabled[spellname]) then
                    showline = showline .. " " .. spellname .. " (" .. FFY_RANK .. " " .. ffy_SpellRanks[spellname] .. ")  "
                else
                    ignoreline = ignoreline .. " " .. spellname .. " (" .. FFY_RANK .. " " .. ffy_SpellRanks[spellname] .. ")  "
                end
            end
        end
        if (showline ~= "") then
            ffy_print(string.gsub(FFY_SHOWLINE, "$l", showline));
        end
        if (ignoreline ~= "") then
            ffy_print(string.gsub(FFY_IGNORELINE, "$l", ignoreline));
        end
        if (ffy_Options.NotifyGroup) then
            ffy_print(FFY_MODE_NOTIFY);
        else
            ffy_print(FFY_MODE_QUIET);
        end
        if ffy_Print_DEBUG then
            ffy_print(FFY_DEBUGON)
        end
        if (ffy_Options.ReplyWhisper == FFY_REPLY_AUTO) then
            ffy_print(FFY_MODE_AUTOREPLY);
        elseif (ffy_Options.ReplyWhisper == FFY_REPLY_FAIL) then
            ffy_print(FFY_MODE_FAILREPLY);
        elseif (ffy_Options.ReplyWhisper == FFY_REPLY_OFF) then
            ffy_print(FFY_MODE_NOREPLY);
        end
        if (ffy_Options.CastOnPets) then
            ffy_print(FFY_PETSMSG_ON);
        else
            ffy_print(FFY_PETSMSG_OFF);
        end
        if (ffy_Options.TellBuffs) then
            ffy_print(FFY_TELLBUFF_ON);
        else
            ffy_print(FFY_TELLBUFF_OFF);
        end
        for playername, spellname in ffy_Paladin_Blessings do
            ffy_print(string.gsub(string.gsub(FFY_WHISPERREQUEST, "$s", spellname), "$t", playername));
        end
        ffy_println(string.gsub(FFY_CONSERVE_MSG, "$m", ffy_Options.ManaConserve));
    
    -- Toggle individual spells notifications
    elseif (command == FFY_CMD_DEBUG) then
        ffy_Print_DEBUG = not ffy_Print_DEBUG;
        if ffy_Print_DEBUG then
            ffy_println(FFY_DEBUGON)
        else
            ffy_println(FFY_DEBUGOFF)
        end
        
    -- Show a list of supported commands
    elseif (command == FFY_CMD_HELP) then
        ffy_println(FFY_HELP_MSG);
        ffy_println(FFY_CMD_TOGGLEUI .. " / " .. FFY_CMD_QUIET .. " / " .. FFY_CMD_TOGGLESPELL .. " / " .. FFY_CMD_SINGLESPELL .. " / " .. FFY_CMD_ALLSPELLS .. " / " .. FFY_CMD_SHOW .. " / " .. FFY_CMD_REPLY .. " / " .. FFY_CMD_PETS .. " / " .. FFY_CMD_TELLBUFF .. " / " .. FFY_CMD_ANNOUNCE .. " / " .. FFY_CMD_BLESS .. " / " .. FFY_CMD_RECAST .. " / " .. FFY_CMD_CONSERVE);
        
    -- Recast all buffs on members of raids
    elseif (command == FFY_CMD_RECAST) then
        ffy_FullRecastMode = true;
        ffy_RecastList = { };
        ffy_println(FFY_RECAST_MSG);
        ffy_Rescan();
    
    -- Show a list of supported commands
    elseif (command == FFY_CMD_PETS) then
        if (args == FFY_ON) then
            ffy_Options.CastOnPets = true;
            ffy_println(FFY_PETSMSG_ON);
        else
            ffy_Options.CastOnPets = false;
            ffy_println(FFY_PETSMSG_OFF);
        end
        ffy_CreateTargetList();
        
    -- Show a list of supported commands
    elseif (command == FFY_CMD_TELLBUFF) then
        if (args == FFY_ON) then
            ffy_Options.TellBuffs = true;
            ffy_println(FFY_TELLBUFF_ON);
        else
            ffy_Options.TellBuffs = false;
            ffy_println(FFY_TELLBUFF_OFF);
        end
        
    -- Announce to the world our supported commands
    elseif (command == FFY_CMD_ANNOUNCE) then
        if (ffy_IsPaladin) then
            local fulllist = ffy_GetBuffList();
            local announcemsg = string.gsub(FFY_ANNOUNCE_MSG, "$l", fulllist);
            ffy_forcespeak(announcemsg);
        else
            ffy_println(FFY_BADANNOUNCE_MSG);
        end
        
    -- Select a blessing for someone manually
    elseif (command == FFY_CMD_BLESS) then
        local arg2;
        p = string.find(args," ");
        if (p ~= nil) then
            arg2 = string.sub(args,p + 1);
            args = string.sub(args,1,p - 1);
            ffy_debug("Bless [" .. args .. "] [" .. arg2 .. "]");
            
            -- Now that we've got our arguments, select the spell and record it
            local spell = ffy_FindSpellByName(arg2);
            if (spell) then
                local settings = ffy_MyData[spell];
                if (ffy_SpellRanks[spell] > 0) then
                    ffy_Paladin_Blessings[string.lower(args)] = spell;
                    ffy_println(string.gsub(string.gsub(FFY_BLESS_MSG, "$s", spell), "$p", args));
                else
                    ffy_println(FFY_BADBLESS_MSG);
                end
            else
                ffy_println(FFY_BADBLESS_MSG);
            end
        else
            ffy_println(FFY_BADBLESS_MSG);
        end
        
    -- Let them know the command syntax
    else
        ffy_errln(FFY_MACRO_ERROR);
    end
end

-------------------------------------------------------------------------------
-- Configure our spellbook
-------------------------------------------------------------------------------
function ffy_Configure()

    -- Check mana conservation settings
    if (ffy_Options.ManaConserve == nil) or (ffy_Options.ManaConserve < 0) then
        ffy_Options.ManaConserve = 0;
    end
    if (ffy_Options.ManaConserve > UnitManaMax("player")) then
        ffy_Options.ManaConserve = UnitManaMax("player");
    end

    -- Fix the rescan timer
    if (ffy_Options.RescanTimer == nil) then
        ffy_Options.RescanTimer = 10.0;
    end
    
    -- Check if we're a paladin
    local playerClass, englishClass = UnitClass("player");
    if (englishClass == FFY_CLASS_PALADIN) then
        ffy_IsPaladin = true;
        ffy_debug("Unit is a paladin.");
    else
        ffy_IsPaladin = false;
        ffy_debug("Unit is NOT a paladin.");
    end
    
    -- first empty out the old spellbook
    if (ffy_MyData == nil) then
        ffy_debug("Setting spellbook for " .. englishClass);
        ffy_MyData = FFY_SPELL_DATA[englishClass];
        if (ffy_MyData == nil) then
            ffy_debug("No spellbook found for " .. englishClass .. ". Abandoning.");
            return;
        end
    end
    
    -- Clear out a few variables
    local s2, s3;
    ffy_SpellID = {};
    ffy_SpellRanks = {};
    for s2, s3 in ffy_MyData do
        ffy_SpellRanks[s2] = 0;
    end
    
    -- Only overwrite preferences if this is our first go
    if (ffy_Options.Version ~= PRAVETZ_FORT) then
        ffy_println(string.gsub(string.gsub(FFY_UPGRADE_MSG, "$v1", ffy_Options.Version), "$v2", PRAVETZ_FORT));
        ffy_Options.Version = PRAVETZ_FORT;
        ffy_Options.Enabled = {};
        
        -- Set some default values
        for s2, s3 in ffy_MyData do
            if (not ffy_Options.Enabled[s2]) then
                ffy_Options.Enabled[s2] = true;
            end
        end
    end

    -- parse through the entire library...
    local i = 1;
    while (true) do
        local spellName, spellRank = GetSpellName(i, BOOKTYPE_SPELL);
        if (not spellName) then
            do break end
        end

        -- Keep track of the highest rank of each spell possessed
		if (ffy_MyData[spellName]) then
            ffy_debug("Spell named [" .. spellName .. "] Rank is [" .. spellRank .. "]");
            
            -- Salvation and Kings have just one rank
            if (spellRank == "") then
                ffy_SpellRanks[spellName] = 1;

            -- Everything else has multiple ranks
            else
                if (spellRank) then
                    local numRank = string.sub(spellRank, string.find(spellRank, " ") + 1);
                    if (ffy_SpellRanks[spellName]) then
            		    ffy_SpellRanks[spellName] = max(numRank, ffy_SpellRanks[spellName]);
                    else
            		    ffy_SpellRanks[spellName] = numRank;
                    end
                    ffy_debug("Highest rank found is [" .. ffy_SpellRanks[spellName] .. "]");
        		end
                spellName = spellName .. "(" .. spellRank .. ")";
            end
		end

        -- put it in the range icon array
        local icon = GetSpellTexture(i, BOOKTYPE_SPELL);
        
        -- add it to the cached ID list
        ffy_debug("Adding " .. spellName);
        ffy_SpellID[spellName] = i;

        i = i + 1
    end
    
    -- Update the button panel
    ffy_Button_Update();
    
    -- Now update all of our icon slots
    ffy_SpellIconSlot = { };
	i = 0;
	for i = FFY_START_SLOT, FFY_END_SLOT do
		if (HasAction(i)) then
            ffy_Tooltip:ClearLines();
            ffy_Tooltip:SetAction(i);
            local slotName = ffy_TooltipTextLeft1:GetText();
    		if (slotName ~= nil) then
    			ffy_debug("Icon Slot " .. i .. " is spell '" .. slotName .. "'.");
    			ffy_SpellIconSlot[slotName] = i;
			end
		end
	end
	return 0;
end

-------------------------------------------------------------------------------
-- Scanning functionalty... this scans the parties and groups
-------------------------------------------------------------------------------
function ffy_Fortify()

    -- First check to see if this person is even capable of casting spells right now
    if (not ffy_SpellID) then
        ffy_errln(FFY_NO_SPELLS);
        return false;
    end

    -- Is the unit mounted?
    if (UnitOnTaxi("player") == 1) then
        ffy_errln(FFY_NO_SPELLS_RDY);
        return false;
    end
    
    -- Keep track of the original target, to switch back later
    local targetName    = "";
    local targetEnemy   = false;
    local casted        = false;
    ffy_People_Missed   = "";

    -- Properly select a target for after we cast a buff
	if (UnitExists("target")) then
		targetName = UnitName("target");
		if (not UnitIsFriend("target", "player")) then
			targetEnemy = true;
		end
	end;

    -- Try to buff everyone in your party or raid
    if (not casted) then
        casted, ffy_LastUnitChecked = ffy_BuffArray(ffy_Unit_Array, ffy_LastUnitChecked);
    end;

    -- Done with buffing.  Restore the original target
    if (targetName == "") then
        ClearTarget();
    else
        if (targetEnemy) then
            TargetLastEnemy();
            ffy_AttackOnDelay = 1.0;
            ffy_debug("Targetting last enemy; attack on delay set to 1.0 secs.");
        else
            TargetByName(targetName);
        end
    end

    -- If we didn't find anything to do, say so to the player's channel
    if (not casted) and (not ffy_Is_Buffing) then
        -- ffy_println(FFY_NOTHING);
    end
    
    -- If we're all buffed up, warn people
    if (not casted) and (ffy_Is_Buffing) then
	    ffy_Button:SetAlpha(0.25);
        ffy_speak(FFY_DONE_BUFFS);
        if (ffy_People_Missed ~= "") then
            ffy_speak(string.gsub(FFY_PEOPLE_MISSED, "$m", strsub(ffy_People_Missed, 2)))
        end
        ffy_LastUnitChecked = 0;
        ffy_LastPetChecked = 0;
        ffy_Is_Buffing = false;
        ffy_FullRecastMode = false;
        ffy_RecastList = { };
    end
    
    -- If we're just starting to cast buffs, warn people
    if (casted) and (not ffy_Is_Buffing) then
	    ffy_Button:SetAlpha(1.0);
        ffy_speak(FFY_STARTING_BUFFS);
        ffy_Is_Buffing = true;
    end
end


-------------------------------------------------------------------------------
-- Check the full array of units and try to buff everyone
-------------------------------------------------------------------------------
function ffy_BuffArray(UnitArray, LastChecked)
    local casted = false;

    -- this is just to save us from going to far
    ffy_debug("Checking array");
    if (LastChecked >= table.getn(UnitArray)) then
        LastChecked = 0;
    end

    if (LastChecked > 0) then
        -- looks like we ended somewhere else.. start at index+1
        local i;

        ffy_debug( "Starting the check at "..(LastChecked+1));
        for i = LastChecked + 1, table.getn(UnitArray) do
            if (ffy_BuffUnit(UnitArray[i])) then
                return true, i;
            end
        end

        ffy_debug( "redoing check from 1... ending at "..LastChecked);
        for i = 1, LastChecked do
            if (ffy_BuffUnit(UnitArray[i])) then
                return true, i;
            end
        end
    else
        local i;
        ffy_debug( "Starting the check at 1");
        for i = 1, table.getn(UnitArray) do
            if (ffy_BuffUnit(UnitArray[i])) then
                return true, i;
            end
        end
    end
    return false, 0;
end

-------------------------------------------------------------------------------
-- Is this unit buffable?
-------------------------------------------------------------------------------
function ffy_OkToBuff(Unit)

    -- Sanity check
    if (Unit == nil) then
        return false;
    end
    
    -- If the unit is bugged, disconnected, or dead, skip it
    if (not UnitExists(Unit)) or (UnitIsDead(Unit)) then
        return false;
    end
    
    -- If the unit is a player and disconnected, skip it
    if (UnitIsPlayer(Unit)) and (not UnitIsConnected(Unit)) then
        return false;
    end
    
    -- Ignore charmed units
    if (UnitIsCharmed(Unit)) then
        return false;
    end

    -- If this isn't a friendly unit, skip it
    if (not UnitIsFriend("player", Unit)) then
        return false;
    end
    
    -- If this person is PVP flagged and we're not, skip them
    if (UnitIsPVP(Unit)) and (not (UnitIsPVP("player"))) then
        return false;
    end
    
    -- Yay!
    return true;
end
    
-------------------------------------------------------------------------------
-- Select which buffs are needed for a unit
-------------------------------------------------------------------------------
function ffy_SelectNeeds(Unit)

    -- Set up the default flags for each person
    local Needs = {};
    local spellname, settings;
    for spellname, settings in ffy_MyData do
        if (ffy_SpellRanks[spellname] > 0) then
            Needs[spellname] = false;

            -- If we're below our mana conserve level, skip this spell
            if (ffy_Options.ManaConserve ~= nil) and (UnitMana("player") - ffy_Options.ManaConserve - ffy_MyData[spellname].manacost[ffy_SpellRanks[spellname]] < 0) then
                -- not enough mana to cast
            
            -- Spells of type "weapon buff" only go on caster if no existing buff is there
            elseif (ffy_MyData[spellname].behavior == FFY_TYPE_WEAPON) then
                if (Unit == "player") then
                    local hasenchant, x1, x2, x3, x4, x5;
                    hasenchant, x1, x2, x3, x4, x5 = GetWeaponEnchantInfo();
                    if (not hasenchant) then
                        Needs[spellname] = true;
                    end
                end

            -- Check spells of type "casters only"
            elseif (ffy_MyData[spellname].behavior == FFY_TYPE_CASTERSONLY) then
                if (UnitPowerType(Unit) == 0) then
                    Needs[spellname] = true;
                end
                
            -- Check spells of type "self buff"
            elseif (ffy_MyData[spellname].behavior == FFY_TYPE_SELF) then
                if (Unit == "player") then
                    Needs[spellname] = true;
                end

            -- All other spells are general buffs
            else
                Needs[spellname] = true;
                ffy_debug(Unit .. " needs " .. spellname);
            end
        end
    end

    -- Figure out what blessing this char wants
    local PreferredBlessing = ffy_Paladin_Blessings[string.lower(UnitName(Unit))];
    local NeedsBlessing = true;
    if (PreferredBlessing) then
        ffy_debug("Unit " .. UnitName(Unit) .. " prefers the blessing " .. PreferredBlessing .. ".");
    else
        ffy_debug("Unit " .. UnitName(Unit) .. " has not chosen a blessing.  Attempting to choose one.");
    end

    -- Check through all buffs on this unit
    ffy_debug( "Buffing - "..Unit);
    local TClass, UClass = UnitClass(Unit);
    for i = 1, FFY_MAXBUFFS do
        local buff_texture = UnitBuff(Unit, i);

        -- If this is an active buff, inspect it
        if buff_texture then
            ffy_TooltipTextRight1:SetText(nil);
            ffy_Tooltip:SetUnitBuff(Unit, i);
            local buff_name = ffy_TooltipTextLeft1:GetText();
            local timetext;

            if (buff_name ~= nil) then
                ffy_debug( buff_name.." found!");
            end

            -- Check if this spell is a buff we should keep track of
            for spellname in Needs do
                if (buff_name == spellname) then
                    
                    -- Special handling code for spell pairs / overrides
                    if (spellname == FFY_SPELL_PRYR_FORTITUDE) or (spellname == FFY_SPELL_FORTITUDE) then
                        Needs[FFY_SPELL_PRYR_FORTITUDE] = false;
                        Needs[FFY_SPELL_FORTITUDE] = false;
                    elseif (spellname == FFY_SPELL_GIFTWILD) or (spellname == FFY_SPELL_MARKWILD) then
                        Needs[FFY_SPELL_GIFTWILD] = false;
                        Needs[FFY_SPELL_MARKWILD] = false;
                    elseif (spellname == FFY_SPELL_ARCANEBRILLIANCE) or (spellname == FFY_SPELL_ARCANEINT) then
                        Needs[FFY_SPELL_ARCANEBRILLIANCE] = false;
                        Needs[FFY_SPELL_ARCANEINT] = false;
                    elseif (spellname == FFY_SPELL_AMPLIFYMAGIC) or (spellname == FFY_SPELL_DAMPENMAGIC) then
                        Needs[FFY_SPELL_AMPLIFYMAGIC] = false;
                        Needs[FFY_SPELL_DAMPENMAGIC] = false;
                    else
                    
                        -- If the user has requested a full recast, do everyone once
                        if (ffy_FullRecastMode) then
                            Needs[spellname] = (ffy_RecastList[UnitName(Unit) .. "-" .. spellname] == nil);
                        
                        -- Otherwise check the recast timer: if the buff is about to expire, recast it
                        else
                            local timer = ffy_NextBuffTime[UnitName(Unit) .. "-" .. spellname];
                            if (timer == nil) then
                                Needs[spellname] = false;
                            else
                                timer = timer - ffy_Clock;
                                ffy_debug("Checking timer " .. timer .. " vs clock " .. ffy_Clock .. " vs recast timer " .. ffy_RecastTimer);
                                if (timer < 0) or (timer > ffy_RecastTimer) then
                                    Needs[spellname] = false;
                                end
                            end
                        end
                    end
                    do break end;
                end
            end
            
            -- Check to see if we've already cast our blessing on this person
            if (buff_name == PreferredBlessing) then
                NeedsBlessing = false;
                ffy_debug("This person already has our blessing " .. PreferredBlessing .. ".");
            end
        end
    end

    -- If they haven't requested something, select a harmless blessing for them
    if (ffy_IsPaladin) and (not PreferredBlessing) then

        -- If this person is a caster, choose wisdom
        local c = string.upper(UnitClass(Unit));
        if (c == FFY_CLASS_DRUID) or (c == FFY_CLASS_PRIEST) or (c == FFY_CLASS_WARLOCK) or (c == FFY_CLASS_MAGE) or (c == FFY_CLASS_SHAMAN) then
            if (Needs[FFY_BLESSING_WISDOM]) and (ffy_SpellRanks[FFY_BLESSING_WISDOM] > 0) then
                PreferredBlessing = FFY_BLESSING_WISDOM;
            elseif (Needs[FFY_BLESSING_SALVATION]) and (ffy_SpellRanks[FFY_BLESSING_SALVATION] > 0) then
                PreferredBlessing = FFY_BLESSING_SALVATION;
            elseif (Needs[FFY_BLESSING_KINGS]) and (ffy_SpellRanks[FFY_BLESSING_KINGS] > 0) then
                PreferredBlessing = FFY_BLESSING_KINGS;
            elseif (Needs[FFY_BLESSING_SANCTUARY]) and (ffy_SpellRanks[FFY_BLESSING_SANCTUARY] > 0) then
                PreferredBlessing = FFY_BLESSING_SANCTUARY;
            elseif (Needs[FFY_BLESSING_LIGHT]) and (ffy_SpellRanks[FFY_BLESSING_LIGHT] > 0) then
                PreferredBlessing = FFY_BLESSING_LIGHT;
            elseif (Needs[FFY_BLESSING_MIGHT]) and (ffy_SpellRanks[FFY_BLESSING_MIGHT] > 0) then
                PreferredBlessing = FFY_BLESSING_MIGHT;
            end

        -- Hunters need wisdom first, then other fighter blessings
        elseif (c == FFY_CLASS_HUNTER) then
            if (Needs[FFY_BLESSING_WISDOM]) and (ffy_SpellRanks[FFY_BLESSING_WISDOM] > 0) then
                PreferredBlessing = FFY_BLESSING_WISDOM;
            elseif (Needs[FFY_BLESSING_KINGS]) and (ffy_SpellRanks[FFY_BLESSING_KINGS] > 0) then
                PreferredBlessing = FFY_BLESSING_KINGS;
            elseif (Needs[FFY_BLESSING_MIGHT]) and (ffy_SpellRanks[FFY_BLESSING_MIGHT] > 0) then
                PreferredBlessing = FFY_BLESSING_MIGHT;
            elseif (Needs[FFY_BLESSING_SANCTUARY]) and (ffy_SpellRanks[FFY_BLESSING_SANCTUARY] > 0) then
                PreferredBlessing = FFY_BLESSING_SANCTUARY;
            elseif (Needs[FFY_BLESSING_LIGHT]) and (ffy_SpellRanks[FFY_BLESSING_LIGHT] > 0) then
                PreferredBlessing = FFY_BLESSING_LIGHT;
            end
        
        -- Otherwise people like Kings the most, then might.
        else
            if (Needs[FFY_BLESSING_KINGS]) and (ffy_SpellRanks[FFY_BLESSING_KINGS] > 0) then
                PreferredBlessing = FFY_BLESSING_KINGS;
            elseif (Needs[FFY_BLESSING_MIGHT]) and (ffy_SpellRanks[FFY_BLESSING_MIGHT] > 0) then
                PreferredBlessing = FFY_BLESSING_MIGHT;
            elseif (Needs[FFY_BLESSING_SANCTUARY]) and (ffy_SpellRanks[FFY_BLESSING_SANCTUARY] > 0) then
                PreferredBlessing = FFY_BLESSING_SANCTUARY;
            elseif (Needs[FFY_BLESSING_LIGHT]) and (ffy_SpellRanks[FFY_BLESSING_LIGHT] > 0) then
                PreferredBlessing = FFY_BLESSING_LIGHT;
            end
        end
        
        -- Set this up as the unit's preferred blessing
        ffy_Paladin_Blessings[string.lower(UnitName(Unit))] = PreferredBlessing;
        ffy_println(string.gsub(string.gsub(FFY_AUTOSELECT_MSG, "$s", PreferredBlessing), "$p", UnitName(Unit)));
    end
    
    -- Quick shortcut for Paladin Blessings
    if (NeedsBlessing) and (PreferredBlessing) then
        ffy_debug("Shortcut to preferred blessing.");
        --Needs = { };
        Needs[PreferredBlessing] = true;
    end

    -- Return what this unit needs
    return Needs;
end

-------------------------------------------------------------------------------
-- Shortcut to test if any buffs are needed at all
-------------------------------------------------------------------------------
function ffy_AnyBuffsToCast()

    -- If no spells available return blank
    if (ffy_MyData == nil) then
        return { };
    end

    -- Check the whole party to see if anyone needs any buffs at the moment
    for i = 1, table.getn(ffy_Unit_Array) do
        if (ffy_OkToBuff(ffy_Unit_Array[i])) then
            local Needs = ffy_SelectNeeds(ffy_Unit_Array[i]);
            for spellname in Needs do
                ffy_debug("Checking " .. UnitName(ffy_Unit_Array[i]) .. " for " .. spellname);
                if (Needs[spellname]) and (ffy_Options.Enabled[spellname]) then
                    ffy_debug("Unit " .. UnitName(ffy_Unit_Array[i]) .. " needs " .. spellname .. " and it is enabled.");
                    return true;
                end
            end
        end
    end
    
    -- Nope, all buffs are up
    return false;
end

-------------------------------------------------------------------------------
-- Cast buffs on a unit
-------------------------------------------------------------------------------
function ffy_BuffUnit(Unit)

    -- Abandon ship if this unit should not be buffed
    if not (ffy_OkToBuff(Unit)) then
        return false;
    end
    
    -- Figure out what buffs this person needs
    local Needs = ffy_SelectNeeds(Unit);

    -- Try to cast a spell on this unit
    local result = false;
    for spellname in Needs do
        if (Needs[spellname]) and (ffy_Options.Enabled[spellname]) then
            ffy_debug("Unit needs " .. spellname .. " and it is enabled.");
        
            -- Check spells of type "weapon buff"
            if (ffy_MyData[spellname].behavior == FFY_TYPE_BLESSING) or (ffy_MyData[spellname].behavior == FFY_TYPE_PARTYBLESSING) then
            
                -- Some blessings are only available to people grouped with you
                ffy_debug("Behavior is " .. ffy_MyData[spellname].behavior);
                if ((ffy_MyData[spellname].behavior == FFY_TYPE_PARTYBLESSING) and (Unit == "target")) then
                    ffy_debug("Can't cast " .. spellname .. " on target; this spell is for group members only.");
                else
                    result = ffy_Cast_Spell(spellname, Unit);
                    ffy_Paladin_Blessings[string.lower(UnitName(Unit))] = spellname;
                    return result;
                end
            
            -- All other spells are general buffs
            else
                ffy_debug(Unit .. " needs " .. spellname .. ".");
                return ffy_Cast_Spell(spellname, Unit);
            end

        -- Buff is not needed
        else
            ffy_debug(Unit .. " doesn't need " .. spellname .. ".");
        end
    end

    return false;
end

-------------------------------------------------------------------------------
-- Cast the appropriate level of a spell on a unit
-------------------------------------------------------------------------------
function ffy_Cast_Spell(spellName, Unit)
    local res = false;
    local unit_level = UnitLevel(Unit);
    local fullspellname;
    local settings = ffy_MyData[spellName];
    
    -- level hide, so set a factice level 999.
    if (unit_level <= 0) then
        unit_level = 999;
    end
    
    -- There is only one rank of these blessings
    if (settings.levels[2] == nil) then
        fullspellname = spellName;

    -- Otherwise, select the appropriate version of the spell
    else
        local selected_rank = 0;
        local i = ffy_SpellRanks[spellName];
        while (i > 0) do
            if (unit_level >= settings.levels[i] - 10) then
                fullspellname = spellName .. "(" .. FFY_RANK .. " " .. i .. ")";
                selected_rank = i;
                do break end;
            end
            i = i - 1;
        end

        -- Maybe nothing is valid
        if (selected_rank < 1) then
            return false;
        end
    end
    
    -- Identify the spell we're casting
    ffy_debug("Starting cast of '" .. fullspellname .. "' on " .. UnitName(Unit) .. ".");
    local spellid = ffy_SpellID[fullspellname];
    
    -- Was this person out of range during the last round?
    if (string.find(ffy_People_Missed, ", " .. UnitName(Unit)) ~= nil) then
        ffy_debug("Person " .. UnitName(Unit) .. " was already missed once.");
        return false;

    -- If not, check them and see if we are in range
    else
        if (not ffy_UnitInRange(Unit, spellName)) then
            ffy_People_Missed = ffy_People_Missed .. ", " .. UnitName(Unit);
            return false;
        end
    end

    -- Final sanity check to make sure the spell exists.
    if (not spellid) then
        ffy_debug("The spell " .. fullspellname .. " was not found.");
        
    -- If we actually have this spell, then yay - give it a shot
    else
        ffy_println( string.gsub( string.gsub(FFY_BUFF_STRING, "$t", UnitName(Unit)), "$s", fullspellname));
        ffy_debug("Spell ID is " .. spellid .. ", book type is " .. SpellBookFrame.bookType);
        TargetByName(UnitName(Unit));
        CastSpell(spellid, SpellBookFrame.bookType);
        res = true;
        
        -- Keep track of when we should next rebuff them with this spell
        ffy_debug("Clock is " .. ffy_Clock .. " and duration is " .. settings.duration[1]);
        ffy_NextBuffTime[UnitName(Unit) .. "-" .. spellName] = ffy_Clock + settings.duration[1];
        if (ffy_FullRecastMode) then
            ffy_RecastList[UnitName(Unit) .. "-" .. spellName] = true;
        end

        -- Optionally, notify the person that we cast on them
        if (ffy_Options.TellBuffs) and (Unit ~= "player") then
            SendChatMessage(FFY_TAG .. string.gsub(FFY_TELLBUFF_MSG, "$s", spellName), "WHISPER", nil, UnitName(Unit));
        end
    end
    
    return res;
end

-------------------------------------------------------------------------------
-- These are updated range functions, hopefully better than the last attempt
-------------------------------------------------------------------------------
function ffy_UnitInRange(Unit, SpellName)

    -- If they're us, we're cool
    if UnitIsUnit(Unit, "player") then
        return true;
    end
    
	-- What a nifty utility function :)
	if (not UnitIsVisible(Unit)) then
		return false;
	end

    -- Try targeting them; if we accidentally get someone else, that's a dead giveaway
	TargetUnit(Unit);
	if not UnitIsUnit("target", Unit) then
	    return false;
	end
	
    -- Check to see if we have this spell in the buttonbar
    local i = ffy_SpellIconSlot[SpellName];
    if (i ~= nil) then
        ffy_debug("Spell " .. SpellName .. " is in icon bar #" .. i .. ", checking range.");
        local result = (IsActionInRange(i) == 1);
        if (result) then
            ffy_debug("In Range");
        else
            ffy_debug("Out of Range");
        end
		return result;
    end
    
	-- Guess that the unit is targetable
	return true;
end

-------------------------------------------------------------------------------
-- Parse through all the events that Fortify needs to know about
-------------------------------------------------------------------------------
function ffy_OnEvent(event)
	if (event == "VARIABLES_LOADED") then
		ffy_Init();
	elseif (event == "SPELLS_CHANGED") then
		ffy_Configure();
	elseif (event == "CHAT_MSG_WHISPER") then
        ffy_CheckWhisper(arg2, arg1);
    elseif (event == "PLAYER_ENTER_COMBAT") then
    	ffy_InCombatMode = true;
        ffy_debug("Now in combat.");
    elseif (event == "PLAYER_LEAVE_COMBAT") then
    	ffy_InCombatMode = false;
        ffy_debug("Leaving combat.");
	end
end

-------------------------------------------------------------------------------
-- Find a spell by checking both its full name and its shortcuts
-------------------------------------------------------------------------------
function ffy_FindSpellByName(label)
    local spell, settings, i;

    -- Only check if we have some spells
    if (ffy_MyData) then
        for spell, settings in ffy_MyData do
            i = 1;
            if (spell) then
                -- If the user went to all the trouble to type it out completely
                if (spell == label) then
                    return spell;
                end
        
                -- Otherwise check all the shortcuts
                while (settings.shortcut[i]) do
                    if (label == settings.shortcut[i]) then
                        return spell;
                    end
                    i = i + 1;
                end
            end
        end
    end
end

-------------------------------------------------------------------------------
-- Show a list of buffs we can cast
-------------------------------------------------------------------------------
function ffy_GetBuffList()
    local fulllist, replymsg;
    local s2, s3;
    for s2, s3 in ffy_MyData do
        if (ffy_SpellRanks[s2] > 0) then
            if (fulllist) then
                fulllist = s2 .. ", " .. fulllist;
            else
                fulllist = s2;
            end
        end
    end
    return fulllist;
end

-------------------------------------------------------------------------------
-- Check to see if this person is requesting a particular blessing
-------------------------------------------------------------------------------
function ffy_CheckWhisper(name, message)

    -- Check to see if this is a recognized whisper
    if (ffy_IsPaladin) then
        if (strlen(message) <= 15) then
            local spell = ffy_FindSpellByName(string.lower(message));
            if (spell) then
                local settings = ffy_MyData[spell];
                ffy_debug("Person '" .. name .. "' wants '" .. spell .. "'.");
                if (ffy_SpellRanks[spell] > 0) then
                    ffy_Paladin_Blessings[string.lower(name)] = spell;
                    
                    -- If the user selected auto reply mode, acknowledge the request
                    if (ffy_Options.ReplyWhisper == FFY_REPLY_AUTO) then
                        ffy_debug("Replying: " .. string.gsub(FFY_RECEIVE_WHISPER, "$s", spell));
                        SendChatMessage(FFY_TAG .. string.gsub(FFY_RECEIVE_WHISPER, "$s", spell), "WHISPER", nil, name);
                    end
                else
                
                    -- If the user selected auto or fail reply mode, let the whisperer know what they can ask for
                    if (ffy_Options.ReplyWhisper == FFY_REPLY_AUTO) or (ffy_Options.ReplyWhisper == FFY_REPLY_FAIL) then
                        local fulllist = ffy_GetBuffList();
                        replymsg = string.gsub(string.gsub(FFY_BAD_WHISPER, "$s", spell), "$l", fulllist);
                        ffy_debug("Replying: " .. replymsg);
                        SendChatMessage(FFY_TAG .. replymsg .. ".", "WHISPER", nil, name);
                    end
                end
            end
        end
    end
end
